/* $Id: fdf_gen.c,v 0.5 2008/03/04 19:02:40 jrothwei Exp jrothwei $ *
** Joseph Rothweiler, Sensicomm LLC, Started 30Jan2008.
** http://www.sensicomm.com
*/

/*
** Copyright 2008 Joseph Rothweiler
**
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Simple program to generate fdf files, for filling out pdf forms.   */

/* The program has been tested using the GCC C compiler under Linux,  **
** but should be compatible with most other C compilers. I compiled   **
** with the command gcc -Wall -o fdf_gen fdf_gen.c                    */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

enum FORMATS {
  DOLLAR_CENTS,
  DOLLAR_CENTS_PAREN,
  STRING,
  STRING3,
  NUMBER,
  BUTTON
} ;
struct FORMAT_TABLE {
	enum FORMATS fmt;
	char *id;
} format_list[] = {
  {DOLLAR_CENTS,        "dollar_cents"},
  {DOLLAR_CENTS_PAREN,  "dollar_cents_paren"},
  {STRING,              "string"},
  {STRING3,             "string3" },
  {NUMBER,              "number" },
  {BUTTON,              "button" },
};
int format_list_len = sizeof(format_list)/sizeof(format_list[0]);

#define FIELDNAMELEN 20  /* Max length, including trailing null. */
#define FIELDTYPELEN  8  /* Max length, including trailing null. */
#define FIELDFDFMAX  12  /* Max length, including trailing null. */
#define MAXFIELDS 200 /* Maximum number of fields in a single pdf file. */
#define FIELDMAXNAMES 10

struct FIELD_STRUCT {
	char name[FIELDNAMELEN];   /* Name used when assiging the value. */
	// char typeid[FIELDTYPELEN]; /* Type of value in the field. */
	enum FORMATS value_type;
	int fdf_count;
	char fdfnames[FIELDMAXNAMES][FIELDFDFMAX]; /* pdf form fields the value goes into. */
} fields[MAXFIELDS];
int numfields;  /* Number of fields actually read from the file. */

int debug = 1;

/******************************************************************************/
int find_field_index(char *id) {
	int k;
	if(debug>10) fprintf(stderr,"Seeking <%s>\n",id);
	for(k=0;k<numfields;k++) {
		if(debug>10) fprintf(stderr,"Comparing [%s]\n",fields[k].name);
		if(strcmp(id,fields[k].name)==0) return k;
	}
	return -1;
}
/******************************************************************************/
enum FORMATS format_lookup(char *id) {
	int k;
	for(k=0;k<format_list_len;k++) {
		if(strcmp(id,format_list[k].id)==0) return format_list[k].fmt;
	}
	return -1;
}
/******************************************************************************
 * A version of fopen that exits on error.                                    */
FILE *Fopen(char *nam, char *mode) {
	FILE *rtn;
	rtn = fopen(nam,mode);
	if(rtn == NULL) {
		fprintf(stderr,"Trouble opening <%s> mode <%s>\n,error %d : %s",
			nam,mode,errno,strerror(errno));
		exit(2);
	}
	return rtn;
}
/******************************************************************************
 * */
void token_add(int numtok, char *toks[],struct FIELD_STRUCT *fld) {
	int k;
	enum FORMATS fmt;
	int need;
	fmt = format_lookup(toks[0]);
	if(((int)fmt)<0) {
		fprintf(stderr,"Bad format %s\n",toks[0]);
		exit(2);
	}
	fld->value_type = fmt;
	strncpy(fld->name,toks[1],FIELDNAMELEN-1);
	fld->name[FIELDNAMELEN-1]='\0';  /* Just for safety. */
	switch(fmt) {
	case DOLLAR_CENTS:
		need = 2;
		break;
	case DOLLAR_CENTS_PAREN:
		need = 2;
		break;
	case STRING:
		need = 1;
		break;
	case STRING3:
		need = 3;
		break;
	case NUMBER:
		need = 1;
		break;
	case BUTTON:
		need = 1;
		break;
	default:
		fprintf(stderr,"Internal error: missed FORMAT %d\n",fmt);
		exit(2);
		break;
	}
	for(k=0;k<FIELDMAXNAMES;k++) {
		if((k+2)>=numtok) break;
		strncpy(&(fld->fdfnames[k][0]),toks[k+2],FIELDFDFMAX-1);
		fld->fdfnames[k][FIELDFDFMAX-1] = '\0';
	}
	fld->fdf_count = k;
	for(;k<FIELDMAXNAMES;k++) {
		fld->fdfnames[k][0] = '\0';
	}
}
/******************************************************************************
 * Read the fields file into the fields structure.
 * The file contains 1 line per value to be placed in the form, formatted as
 * type name count fdf_names ...  */
void fields_read(
  FILE *fil
) {
	int linecount;
	numfields = 0;
	for(linecount=0;;linecount++) {
		char inlin[200];
#define MAXTOKENS (FIELDMAXNAMES+3)
		char *tokens[MAXTOKENS];
		int tokptr;
		int intoken;
		int k;
		/* Read the next line. */
		if(fgets(inlin,200,fil)==NULL) break;
		/* Split into individual tokens. */
		tokptr = 0;
		intoken = 0;
		if(inlin[0]=='#') {
			if(debug>10) {
				fprintf(stderr,"Comment: %s\n",inlin);
			}
			continue;
		}
		for(k=0;k<200;k++) {
			if(inlin[k] == '\0') {
				if(debug>10) fprintf(stderr,"End line at %d\n",k);
				break;
			}
			if(intoken) {
				if(isspace(inlin[k])) {
					inlin[k] = '\0';
					intoken = 0;
				}
			} else {
				if(isspace(inlin[k])) {
					inlin[k] = '\0';
				} else {
					if(debug>10) fprintf(stderr,"Start token at %d\n",k);

					if(tokptr >= MAXTOKENS) {
						fprintf(stderr,"Too many tokens on line %d.\n",
							linecount+1);
						exit(2);
					}
					tokens[tokptr++] = inlin+k;
					intoken = 1;
				}
			}
		}
		for(;k<200;k++) inlin[k] = '\0';
		if(debug>10) {
			int k;
			fprintf(stderr,"Got %d tokens:",tokptr);
			for(k=0;k<tokptr;k++) fprintf(stderr," <%s>",tokens[k]);
			fprintf(stderr,"\n");
		}
		/* Put the tokens into the structure. */
		if(numfields>=MAXFIELDS) {
			fprintf(stderr,"Too many fields. Increase MAXFIELDS and recompile\n");
			exit(2);
		}
		token_add(tokptr,tokens,fields+numfields);
		numfields++;
	}
}
/******************************************************************************
 * */
void values_process(FILE *inf, FILE *outf) {
	int linecount;
	for(linecount=0;;linecount++) {
		char inlin[200];
		int len;
		int k;
		char *idptr, *valptr;
		int field_idx;

		/* Read the next line. */

		if(fgets(inlin,200,inf)==NULL) break;
		if(debug>10) fprintf(stderr,"inlin[%d] <%s>\n",strlen(inlin),inlin);

		/* Clean the newline off the end. */

		len=strlen(inlin);
		for(len--;len>=0;len--) {
			if(inlin[len]=='\n') inlin[len]='\0';
			else if(inlin[len]=='\r') inlin[len]='\0';
			else break;
		}
		if(debug>10) fprintf(stderr,"Cleaned inlin[%d] <%s>\n",strlen(inlin),inlin);

		/* Find the ID (starts with first nonzero character). */

		len=strlen(inlin);
		if(len == 0) {
			if(debug>10) fprintf(stderr,"Ignoring blank line %d\n",linecount+1);
			continue;   /* Ignore blank lines. */
		}
		for(k=0;k<len;k++) {
			if(inlin[k]!=' ') break;
		}
		if(k<len) {
			idptr=inlin+k;
		} else {
			fprintf(stderr,"No ID found on line %d\n",linecount+1);
			exit(2);
		}
		if(*idptr == '#') {
			if(debug>10) fprintf(stderr,"Comment line %d: <%s>\n",linecount+1,inlin);
			continue;  /* Ignore comment lines. */
		}
		/* Find the space between id and value. */

		for(;k<len;k++) {
			if(inlin[k]==' ') {
				inlin[k]='\0';
				break;
			}
		}
		if(k<len-1) {
			valptr = inlin+k+1;
		} else {
			fprintf(stderr,"No value found on line %d\n",linecount+1);
			exit(2);
		}
		if(debug>10) {
			fprintf(stderr,"Line %d: <%s> <%s>\n",linecount+1,idptr,valptr);
		}

		/***********************
		 * We've got the id and value strings. Now convert them into
		 * the proper format for the fdf file. */

		field_idx = find_field_index(idptr);
		if(field_idx < 0) {
			fprintf(stderr,"Unknown id at line %d: <%s>\n",linecount+1,idptr);
			exit(2);
		}
		/* Process it, depending on the type. */
		switch(fields[field_idx].value_type) {
		char ch, *vp;
		int usedfields;
		double dval;
		int val_dol,val_cent;
		case DOLLAR_CENTS:
			if(fields[field_idx].fdf_count!=2) {
				fprintf(stderr,"Bad field count at line %d\n",linecount+1);
				exit(2);
			}
			/* FIXME: probably better to do this in integers for exactness. */
			sscanf(valptr,"%lg",&dval);
			val_cent=floor(fabs(dval)*100.0+0.001);
			val_dol=val_cent/100;
			val_cent %= 100;
			if(dval<0.0) val_dol *= -1;
			fprintf(outf,"<</T(%s)/V(%d)>>\n",
				&(fields[field_idx].fdfnames[0][0]),val_dol);
			fprintf(outf,"<</T(%s)/V(%2.2d)>>\n",
				&(fields[field_idx].fdfnames[1][0]),val_cent);
			break;
		case DOLLAR_CENTS_PAREN:
			if(fields[field_idx].fdf_count!=2) {
				fprintf(stderr,"Bad field count at line %d\n",linecount+1);
				exit(2);
			}
			/* FIXME: probably better to do this in integers for exactness. */
			sscanf(valptr,"%lg",&dval);
			val_cent=floor(fabs(dval)*100.0+0.001);
			val_dol=val_cent/100;
			val_cent %= 100;
			if(dval>0.0) {
				fprintf(outf,"<</T(%s)/V(%d)>>\n",
					&(fields[field_idx].fdfnames[0][0]),val_dol);
				fprintf(outf,"<</T(%s)/V(%2.2d)>>\n",
					&(fields[field_idx].fdfnames[1][0]),val_cent);
			} else {
				fprintf(outf,"<</T(%s)/V(\\(%d)>>\n",
					&(fields[field_idx].fdfnames[0][0]),val_dol);
				fprintf(outf,"<</T(%s)/V(%2.2d\\))>>\n",
					&(fields[field_idx].fdfnames[1][0]),val_cent);
			}
			break;
		case STRING3:
			vp = valptr;
			ch = *vp++;   /* First character defines the separator. */
			usedfields = 0;
			for(k=1;;k++) {
				int flag;
				flag = valptr[k]=='\0';
				if( (valptr[k]=='\0') ||(valptr[k]==ch) ) {
					valptr[k]='\0'; /* FIXME: modifies the value. */
					if(usedfields >= fields[field_idx].fdf_count) {
						fprintf(stderr,
							"Too many fields in string3 at line %d: <%s>\n",
							linecount+1,idptr);
						exit(2);
					}
					fprintf(outf,"<</T(%s)/V(%s)>>\n",
						&(fields[field_idx].fdfnames[usedfields][0]),vp);
					vp = valptr+k+1;
					usedfields++;
				}
				if(flag) break;
			}
			if(usedfields != fields[field_idx].fdf_count) {
				fprintf(stderr,
					"Not all %d fields used in string3 at line %d: <%s>\n",
					fields[field_idx].fdf_count,
					linecount+1,idptr);
			}
			break;
		case STRING:   /* String and number are really handled the same. */
		case BUTTON:   /* Button is just a string with a list of allowed values. */
		               /* I don't check the values yet. */
		case NUMBER:
			if(fields[field_idx].fdf_count!=1) {
				fprintf(stderr,"Bad field count at line %d\n",linecount+1);
				exit(2);
			}
			fprintf(outf,"<</T(%s)/V(%s)>>\n",
				&(fields[field_idx].fdfnames[0][0]),valptr);
			break;
		default:
			fprintf(stderr,"Internal error: missed a FORMAT %d\n",
				fields[field_idx].value_type );
			exit(2);
			break;
		}
	}
}
/******************************************************************************/
char header[] = 
  "%FDF-1.2\n"
  "1 0 obj<</FDF<< /Fields[\n" ;
char trailer[] = 
  "] >> >>\n"
  "endobj\n"
  "trailer\n"
  "<</Root 1 0 R>>\n"
  "%%EOF\n";
/******************************************************************************
 * */
int main(int argc, char *argv[]) {
	FILE *field_file, *value_file, *out_file;
	if(argc!=4) {
		fprintf(stderr,"cmd: %s fields values out.fdf\n",argv[0]);
		fprintf(stderr," fields: file containing the field descriptions.\n");
		fprintf(stderr," values: file containing the values for each field.\n");
		fprintf(stderr," out.fdf: The output file (will be overwritten if it exists).\n");
		exit(2);
	}

	field_file = Fopen(argv[1],"r");
	value_file = Fopen(argv[2],"r");
	out_file   = Fopen(argv[3],"w");

	fields_read(field_file);
	fclose(field_file);

	fprintf(out_file,"%s",header);
	values_process(value_file,out_file);
	fprintf(out_file,"%s",trailer);

	fclose(value_file);
	fclose(out_file);

	return(0);
}
